﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Drawing.Drawing2D;

namespace StockParser
{
    struct StockInfo
    {
        public DateTime time;
        public double open;
        public double close;
        public double high;
        public double low;
        public Rectangle drawRect; //股票所画的矩形的位置
        public int drawMid; //股票画图时中间的位置
        public int drawCloseY; //收盘时的Y坐标
    }

    public partial class StockChart : UserControl
    {
        private Color blueColor;
        private Color redColor;
        private Color redColor2;
        private Pen bluePen;
        private Pen redPen;
        private Pen whitePen;
        private Pen redDotPen;
        private SolidBrush purpleBrush;
        private SolidBrush blueBrush;
        private SolidBrush redBrush;
        private SolidBrush grayBrush;
        private SolidBrush whiteBrush;
        private SolidBrush yellowBrush;
        private int m_iCurRectWidth = 0; //当前矩形宽度
        private int m_iCurSpaceWidth = 3; //当前矩形间距
        private int m_iMargin = 20; //左右上下留的空白距离
        private int m_iBottomTextSpace = 50; //下面留下的文本空白
        private int m_iRightTextSpace = 50; //右侧留下的文本空白
        private float m_fCurMax = 0f; //k表最大值
        private float m_fCurMin = 0f; //k表最小值
        private double m_dCurYScale = 0d; //Y轴方向的比例 (可画的高度-2*margin)/(最大值-最小值)
        private int[,] m_arrDrawInfo = new int[,] { { 1, 0 }, { 1, 2 }, { 3, 2 }, { 7, 3 }, { 10, 4 }, { 15, 4 }, { 20, 4 }, { 25, 4 }, { 40, 4 } };
        private int m_iDefaultScaleLevel = 7;
        private int m_iCurScaleLevel = 0; //当前缩放等级
        private int m_iMinGridHeight = 70; //水平虚线格子之间的最小高度
        private int m_iMaxGridHeight = 95; //水平虚线格子之间的最大高度
        private int m_iStockIndex = 0; //十在线所在的位置的股票在listStock中的索引
        private Rectangle rectPanel; //当前的this面板
        private DataTable dt; //当前数据表
        private Bitmap m_bmpBuffer; //缓冲图片
        private bool m_bDrawCross = false; //画十字线及线两端的指示
        private bool m_bZoomCenterChange = false; //放大缩小图时, 要不要以当根据当前的的选中来放大
        private List<StockInfo> m_listStock = new List<StockInfo>();
        private Point m_ptLastMouse; //上次鼠标所在位置
        private int m_iMousePointSameTimes = 0; //定时器统计到鼠标与上一次位置相同位置的次数
        private int m_iDelay = 5; //延迟多长时间显示Stock的信息 5*100ms
        public StockChart()
        {
            blueColor = Color.FromArgb(84, 252, 252);
            redColor = Color.FromArgb(255, 50, 50);
            redColor2 = Color.FromArgb(128, 0, 0);
            bluePen = new Pen(blueColor);
            redPen = new Pen(redColor);
            whitePen = new Pen(Color.White);
            redDotPen = new Pen(redColor);
            purpleBrush = new SolidBrush(Color.FromArgb(64, 0, 128));
            blueBrush = new SolidBrush(blueColor);
            redBrush = new SolidBrush(redColor);
            whiteBrush = new SolidBrush(Color.White);
            yellowBrush = new SolidBrush(Color.Yellow);
            grayBrush = new SolidBrush(Color.FromArgb(57, 57, 57));
            InitializeComponent();
            // Set Optimized Double Buffer to reduce flickering
            this.SetStyle(ControlStyles.UserPaint, true);
            this.SetStyle(ControlStyles.AllPaintingInWmPaint, true);
            this.SetStyle(ControlStyles.OptimizedDoubleBuffer, true);

            // Redraw when resized
            this.SetStyle(ControlStyles.ResizeRedraw, true);
            m_iCurScaleLevel = m_iDefaultScaleLevel;
        }

        /// <summary>
        /// 初始化从数据库中读取股票数据
        /// </summary>
        /// <param name="strTime">最大或最小时间字符串, 如果为空字符串, 则从最后的时间取数据</param>
        /// <param name="bIsMinTime">是否是最小时间</param>
        private void InitData(string strTime = "", bool bIsMinTime = true)
        {
            m_iCurRectWidth = m_arrDrawInfo[m_iCurScaleLevel, 0] - 2 > 0 ? m_arrDrawInfo[m_iCurScaleLevel, 0] - 2 : 1; //矩形宽度最少为3的, 否则为1
            m_iCurSpaceWidth = m_arrDrawInfo[m_iCurScaleLevel, 1];
            rectPanel = this.Bounds;
            rectPanel.Size = new Size(rectPanel.Size.Width - m_iRightTextSpace, rectPanel.Size.Height - m_iBottomTextSpace);

            //可画区域 = 左右宽度 - 2 * margin
            //能画多少个矩形
            int iCanDrawRectCount = (int)Math.Floor((double)(rectPanel.Width - m_iMargin * 2) / (m_iCurRectWidth + m_iCurSpaceWidth));
            string strSql = "select top " + iCanDrawRectCount
                + " * from stockdata where stockcode = 'f010' {condition}";
            bool bReverseDataTable = true; //下面读取出的数据要不要反排序
            if (strTime != "")
            {
                if (bIsMinTime)
                {
                    object objMinTime = DbHelperSQL.GetSingle("select top 1 Stocktime from stockdata where StockTime < '"
                        + strTime + "' order by stocktime desc");
                    if (objMinTime != null)
                    {
                        strSql = strSql.Replace("{condition}", " and Stocktime >= '" + objMinTime.ToString()
                            + "' order by stocktime asc");
                        bReverseDataTable = false;
                    }
                    else
                        return;
                }
                else
                {
                    object objMaxTime = DbHelperSQL.GetSingle("select top 1 Stocktime from stockdata where StockTime > '"
                        + strTime + "' order by stocktime asc");
                    if (objMaxTime != null)
                    {
                        strSql = strSql.Replace("{condition}", " and StockTime <= '"
                            + objMaxTime.ToString() + "' order by stocktime desc");
                    }
                    else
                        return;
                }
            }
            else
            {
                strSql = strSql.Replace("{condition}", " order by stocktime desc");
            }
            StockInfo stock = new StockInfo();
            try
            {
                if (m_bZoomCenterChange)
                {
                    stock = m_listStock[m_iStockIndex]; //当前的stock
                    //如果stock找到, 则查找他右边的 iCanDrawRectCount/2个元素
                    int iRight = (int)Math.Round(iCanDrawRectCount * 1.0 / 2);
                    string strSqlQuery = "select top "
                        + iRight + " stocktime from stockdata where stockcode = 'f010' and stocktime >= '"
                        + stock.time.ToString() + "' order by stocktime asc";
                    DataTable dtRight = DbHelperSQL.Query(strSqlQuery).Tables[0];
                    object obj = dtRight.Rows[dtRight.Rows.Count - 1][0];
                    if (obj != null)
                    {
                        strSql = "select top " + iCanDrawRectCount
                            + " * from stockdata where stockcode = 'f010'and stocktime <='"
                            + obj.ToString() + "' order by stocktime desc";
                    }
                }
            }
            catch (System.Exception ex)
            {

            }

            DataSet ds = DbHelperSQL.Query(strSql);
            if (ds != null && ds.Tables.Count > 0 && ds.Tables[0].Rows.Count > 0)
            {
                if (bReverseDataTable)
                {
                    //重按时间排序
                    DataRow[] rows = ds.Tables[0].Select("", "Stocktime asc");
                    dt = ds.Tables[0].Clone();
                    dt.Clear();
                    foreach (DataRow row in rows)
                        dt.ImportRow(row);
                }
                else
                {
                    dt = ds.Tables[0];
                }
                //找出最大值, 最小值
                m_fCurMax = (float)dt.AsEnumerable().Select(t => t.Field<double>("High")).Max();
                m_fCurMin = (float)dt.AsEnumerable().Select(t => t.Field<double>("Low")).Min();
                m_dCurYScale = (rectPanel.Height - 2 * m_iMargin) / (m_fCurMax - m_fCurMin);
                if (stock.drawMid != 0)
                {
                    for (int i = 0; i < dt.Rows.Count; ++i)
                    {
                        try
                        {
                            if ((DateTime)dt.Rows[i]["Stocktime"] == stock.time)
                            {
                                m_iStockIndex = i;
                            }
                        }
                        catch (System.Exception ex)
                        {

                        }

                    }
                }
            }
            else
                dt = null;
            m_bZoomCenterChange = false;
        }

        /// Override OnPaint method
        protected override void OnPaint(PaintEventArgs e)
        {
            base.OnPaint(e);
            DrawStock(e.Graphics);
        }
        /// <summary>
        /// 画图的主函数
        /// </summary>
        /// <param name="tg"></param>
        private void DrawStock(Graphics tg)
        {
            m_bmpBuffer = new Bitmap(this.Width, this.Height);
            Graphics g = Graphics.FromImage(m_bmpBuffer);
            //g.Clear(Color.Black);
            g.SmoothingMode = SmoothingMode.Default;
            g.PixelOffsetMode = System.Drawing.Drawing2D.PixelOffsetMode.Default;
            if (dt != null)
            {
                m_listStock.Clear();
                int iArrayLen = m_arrDrawInfo.Length / m_arrDrawInfo.Rank;
                for (int i = 0; i < dt.Rows.Count; ++i)
                {
                    //画个矩形
                    double open = ((double)dt.Rows[i]["Open"] - m_fCurMin) * m_dCurYScale;
                    double close = ((double)dt.Rows[i]["Close"] - m_fCurMin) * m_dCurYScale;
                    double high = ((double)dt.Rows[i]["High"] - m_fCurMin) * m_dCurYScale;
                    double low = ((double)dt.Rows[i]["Low"] - m_fCurMin) * m_dCurYScale;
                    int iRectLeft = (int)(m_iMargin + i * (m_iCurRectWidth + m_iCurSpaceWidth));
                    int iRectTop = (int)(Math.Max(open, close) + m_fCurMin) + m_iBottomTextSpace;
                    int iRectMid = (iRectLeft * 2 + m_iCurRectWidth) / 2; //矩形中间点横坐标
                    int iHigh = (int)(high + m_fCurMin) + m_iBottomTextSpace; //最高点纵坐标
                    int iLow = (int)(low + m_fCurMin) + m_iBottomTextSpace;//最低点纵坐标
                    int iRectHeight = (int)Math.Abs(open - close); //矩形的高度
                    int iRectBottom = (int)(Math.Min(open, close) + m_fCurMin) + m_iBottomTextSpace; //矩形的底坐标
                    Rectangle rect = new Rectangle(iRectLeft, iRectTop, (int)m_iCurRectWidth, iRectHeight);
                    rect = ConvertRectCoordinate(this.Height, rect);
                    Pen curPen = redPen;
                    if (open > close)
                    {
                        curPen = bluePen;
                        g.FillRectangle(blueBrush, rect);
                    }
                    g.DrawRectangle(curPen, rect);
                    //画线
                    //画最高线
                    if (iHigh != iRectTop)
                        g.DrawLine(curPen, iRectMid, this.Height - iHigh, iRectMid, this.Height - iRectTop);
                    //画最低线
                    if (iLow != iRectBottom)
                        g.DrawLine(curPen, iRectMid, this.Height - iLow, iRectMid, this.Height - iRectBottom);
                    StockInfo stockInfo = new StockInfo();
                    stockInfo.open = (double)dt.Rows[i]["Open"];
                    stockInfo.close = (double)dt.Rows[i]["Close"];
                    stockInfo.high = (double)dt.Rows[i]["High"];
                    stockInfo.low = (double)dt.Rows[i]["Low"];
                    stockInfo.time = DateTime.Parse(dt.Rows[i]["StockTime"].ToString());
                    stockInfo.drawRect = rect;
                    stockInfo.drawMid = iRectMid;
                    stockInfo.drawCloseY = iHigh;
                    if (open > close)
                        stockInfo.drawCloseY = iLow;
                    m_listStock.Add(stockInfo);
                }
                //画坐标轴
                //横坐标线
                g.DrawLine(redPen, 0, rectPanel.Height, this.Width, rectPanel.Height);
                //纵坐标线
                g.DrawLine(redPen, rectPanel.Width, 0, rectPanel.Width, this.Height);
                //画坐标
                //横坐标
                //最后两个比例画天, 其它有空格的画月, 其它的就画年了
                if (m_iCurScaleLevel > iArrayLen - 3)
                {
                    //画每一天
                    foreach (StockInfo stock in m_listStock)
                    {
                        //画小竖线
                        g.DrawLine(redPen, stock.drawMid, rectPanel.Height, stock.drawMid, rectPanel.Height + 5);
                        //画天文字
                        g.DrawString(stock.time.Day.ToString(), this.Font, redBrush, stock.drawMid + 2, rectPanel.Height + 2);
                    }
                }
                else if (m_arrDrawInfo[m_iCurScaleLevel, 1] > 0)
                {
                    //画每个月
                    foreach (StockInfo stock in m_listStock)
                    {
                        if (stock.time.Day == 1)
                        {
                            //画小竖线
                            g.DrawLine(redPen, stock.drawMid, rectPanel.Height, stock.drawMid, rectPanel.Height + 5);
                            //画天文字
                            g.DrawString(stock.time.Month.ToString(), this.Font, redBrush, stock.drawMid + 2, rectPanel.Height + 2);
                        }
                    }
                }
                else
                {
                    //画每年
                    foreach (StockInfo stock in m_listStock)
                    {
                        if (stock.time.Month == 1 && stock.time.Day == 1)
                        {
                            //画小竖线
                            g.DrawLine(redPen, stock.drawMid, rectPanel.Height, stock.drawMid, rectPanel.Height + 5);
                            //画天文字
                            g.DrawString(stock.time.Year.ToString(), this.Font, redBrush, stock.drawMid + 2, rectPanel.Height + 2);
                        }
                    }
                }
                //画纵坐标 区格高度为m_iMinGridHeight(70)-m_iMaxGridHeight(95)
                int iGridCountMin = (int)Math.Ceiling(rectPanel.Height * 1.0 / m_iMaxGridHeight);
                int iGridCountMax = (int)Math.Floor(rectPanel.Height * 1.0 / m_iMinGridHeight);
                int iGridCount = iGridCountMin;
                if (iGridCountMax - iGridCountMin > 1)
                {
                    iGridCount = (int)Math.Round((iGridCountMax + iGridCountMin) * 1.0 / 2);
                }
                int iGridHeight = (int)Math.Round(rectPanel.Height * 1.0 / iGridCount);
                for (int i = 1; i < iGridCount; ++i)
                {
                    int iDrawHeight = rectPanel.Height - iGridHeight * i;
                    g.DrawLine(redDotPen, 0, iDrawHeight, rectPanel.Width, iDrawHeight);
                    //写文字
                    int iStockValue = (int)Math.Round(iGridHeight * i / m_dCurYScale + m_fCurMin);
                    g.DrawString(iStockValue.ToString(), this.Font, redBrush, rectPanel.Width + 1, iDrawHeight - 2);
                }
                Point pMouse = this.PointToClient(Control.MousePosition); //鼠标的位置
                //查找十字线的横坐时, 可以根据鼠标点击位置, 计算出在画面板横坐标的比例, 然后用这个比例到listStock下面的索引查找对应的stock
                //要检测下整年下面是否正确, 如果不正确, 可以向前或向后查找细查
                //画十字线
                if (m_bDrawCross)
                {
                    StockInfo stock = m_listStock[m_iStockIndex];
                    //竖线
                    g.DrawLine(whitePen, stock.drawMid, 0, stock.drawMid, this.Height - m_iBottomTextSpace);
                    //竖线日期
                    Rectangle rectTipBottom = new Rectangle(stock.drawMid, this.Height - m_iBottomTextSpace, 85, 20);
                    g.FillRectangle(purpleBrush, rectTipBottom);
                    g.DrawRectangle(redPen, rectTipBottom);
                    g.DrawString(stock.time.ToString("yyyy-MM-dd") + "," + "日一二三四五六".Substring((int)stock.time.DayOfWeek,1),
                        this.Font, redBrush, stock.drawMid + 2, this.Height - m_iBottomTextSpace + 5);
                    //横线
                    g.DrawLine(whitePen, 0, pMouse.Y, this.Width - m_iRightTextSpace, pMouse.Y);
                    //横线右边的值
                    int iYValue = (int)Math.Round((this.Height - pMouse.Y - m_iBottomTextSpace) / m_dCurYScale + m_fCurMin);
                    Rectangle rectTipRight = new Rectangle(this.Width - m_iRightTextSpace, pMouse.Y - 10, m_iRightTextSpace, 20);
                    g.FillRectangle(purpleBrush, rectTipRight);
                    g.DrawRectangle(redPen, rectTipRight);
                    g.DrawString(iYValue.ToString(), this.Font, redBrush, this.Width - m_iRightTextSpace + 2, pMouse.Y - 5);
                }
                if (m_iMousePointSameTimes >= m_iDelay)
                {
                    //画鼠标位置的股票提示
                    int iStockIndex = (int)Math.Floor((pMouse.X - m_iMargin) * 1.0 / (m_iCurRectWidth + m_iCurSpaceWidth));
                    try
                    {
                        StockInfo curPtStock = m_listStock[iStockIndex];
                        if (curPtStock.drawRect.Contains(pMouse))
                        {
                        
                            //矩形包含点
                            int rectX = pMouse.X + 10;
                            int rectY = pMouse.Y + 10;
                            Rectangle rectStockTip = new Rectangle(pMouse.X + 10, pMouse.Y + 10, 90, 80);
                            g.FillRectangle(grayBrush, rectStockTip);
                            Font boldFont = new Font(this.Font, FontStyle.Bold);
                            g.DrawString(curPtStock.time.ToString("yyyy-MM-dd") + "," 
                                + "日一二三四五六".Substring((int)curPtStock.time.DayOfWeek, 1),
                                boldFont, yellowBrush, rectX + 2, rectY + 5);
                            g.DrawString("开:" + curPtStock.open.ToString(), this.Font, whiteBrush, rectX + 2, rectY + 5 + 15);
                            g.DrawString("高:" + curPtStock.high.ToString(), this.Font, whiteBrush, rectX + 2, rectY + 5 + 15 * 2);
                            g.DrawString("低:" + curPtStock.low.ToString(), this.Font, whiteBrush, rectX + 2, rectY + 5 + 15 * 3);
                            g.DrawString("收:" + curPtStock.close.ToString(), this.Font, whiteBrush, rectX + 2, rectY + 5 + 15 * 4);
                        }
                    }
                    catch (System.Exception ex)
                    {
                        //鼠标移出去会有异常, 所以这儿捕下
                    }
                }
            }
            tg.DrawImage(m_bmpBuffer, 0, 0);　　//把画布贴到画面上
        }

        //右下角为坐标原点的坐标转换成左上角为坐标原点的坐标
        private Rectangle ConvertRectCoordinate(int height, Rectangle rectOrign)
        {
            Rectangle rect = new Rectangle(rectOrign.Location, rectOrign.Size);
            rect.Y = height - rect.Location.Y;
            return rect;
        }

        protected override void OnResize(EventArgs e)
        {
            base.OnResize(e);
            m_bDrawCross = false;//变换尺寸不画十字线
            ReDrawStock();
        }

        /// <summary>
        /// 重画股票
        /// </summary>
        /// <param name="bInitData">是否要重置数据</param>
        private void ReDrawStock(bool bInitData = true)
        {
            if (bInitData)
                InitData();
            Invalidate();
        }
        /// <summary>
        /// 鼠标移动事件
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void StockChart_MouseMove(object sender, MouseEventArgs e)
        {
            if (m_listStock.Count == 0)
            {
                return;
            }
            Point pMouse = this.PointToClient(Control.MousePosition);
            int iStockIndex = (int)Math.Floor((pMouse.X - m_iMargin) * 1.0 / (m_iCurRectWidth + m_iCurSpaceWidth));
            if(m_iMousePointSameTimes >= m_iDelay)
                SetStockIndex(iStockIndex, true);
            else
                SetStockIndex(iStockIndex);
        }
        /// <summary>
        /// 设置当前竖线在列表中的索引
        /// </summary>
        /// <param name="iStockIndex"></param>
        private void SetStockIndex(int iStockIndex, bool bForceRedraw = false)
        {
            if (iStockIndex < 0)
                iStockIndex = 0;
            else if (iStockIndex > m_listStock.Count - 1)
                iStockIndex = m_listStock.Count - 1;
            m_iStockIndex = iStockIndex;
            if (m_bDrawCross || bForceRedraw)
                ReDrawStock(false);
        }
        private void StockChart_Click(object sender, EventArgs e)
        {
            m_bDrawCross = true;
            ReDrawStock(false);
        }
        //父窗口调用的按键事件
        public void ProcessKeyDown(Keys key)
        {
            int iLen = m_arrDrawInfo.Length / m_arrDrawInfo.Rank;
            if (key == Keys.Up)
            {
                if (m_bDrawCross)
                    m_bZoomCenterChange = true;
                m_iCurScaleLevel++;
                if (m_iCurScaleLevel > iLen - 1)
                    m_iCurScaleLevel = iLen - 1;
                ReDrawStock();
            }
            else if (key == Keys.Down)
            {
                if (m_bDrawCross)
                    m_bZoomCenterChange = true;
                m_iCurScaleLevel--;
                if (m_iCurScaleLevel < 0)
                    m_iCurScaleLevel = 0;
                ReDrawStock();
            }
            else if (key == Keys.Left)
            {
                if (m_iStockIndex == 0)
                {
                    InitData(m_listStock[0].time.ToString(), true);
                    Invalidate();
                }
                else
                {
                    int iStockIndex = m_iStockIndex;
                    SetStockIndex(--iStockIndex);
                }
            }
            else if (key == Keys.Right)
            {
                if (m_iStockIndex == m_listStock.Count - 1)
                {
                    InitData(m_listStock[m_listStock.Count - 1].time.ToString(), false);
                    Invalidate();
                }
                else
                {
                    int iStockIndex = m_iStockIndex;
                    SetStockIndex(++iStockIndex);
                }
            }
        }
        /// <summary>
        /// 用来监视鼠标停放1秒钟是否要画一个当前股票的信息
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void timerWatch_Tick(object sender, EventArgs e)
        {
            Point pMouse = this.PointToClient(Control.MousePosition);
            if (m_ptLastMouse != pMouse)
            {
                m_ptLastMouse = pMouse;
                m_iMousePointSameTimes = 0;
            }
            else
            {
                if (m_iMousePointSameTimes < m_iDelay)
                {
                    ++m_iMousePointSameTimes;
                    ReDrawStock(false);
                }
            }
        }
    }
}
